/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache;

import org.jboss.cache.config.Option;
import org.jboss.cache.lock.NodeLock;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.transaction.GlobalTransaction;

import javax.transaction.Transaction;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * This context holds information specific to a method invocation.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class InvocationContext implements Cloneable
{
   private Transaction transaction;
   private GlobalTransaction globalTransaction;
   private Option optionOverrides;
   // defaults to true.
   private boolean originLocal = true;
   private boolean txHasMods;
   private boolean cacheLoaderHasMods;
   private boolean localRollbackOnly;
   private MethodCall methodCall;

   // used to store cache peeks within the scope of a single context. Performing a cache peek can be a huge bottle neck.
   // See JBCACHE-811
   // backing out for now
   //private Map<Fqn, NodeSPI> peekedNodes = new HashMap<Fqn, NodeSPI>();

   /**
    * Retrieves a node that may have previously been "peeked" within the scope of the same invocation.
    *
    * @param f fqn of node to find
    * @return node, if previously peeked, or null if not.
    * @since 2.1.0
    */
//   public NodeSPI getPeekedNode(Fqn f)
//   {
//      return peekedNodes.get(f);
//   }

   /**
    * Adds a node to the previously peeked list.
    *
    * @param n node to add
    * @param f fqn of node
    * @since 2.1.0
    */
//   public void savePeekedNode(NodeSPI n, Fqn f)
//   {
//      peekedNodes.put(f, n);
//   }

   /**
    * Wipe list of previously peeked nodes.
    *
    * @since 2.1.0
    */
//   public void wipePeekedNodes()
//   {
//      peekedNodes.clear();
//   }
   public void setLocalRollbackOnly(boolean localRollbackOnly)
   {
      this.localRollbackOnly = localRollbackOnly;
   }

   /**
    * Retrieves the transaction associated with this invocation
    *
    * @return The transaction associated with this invocation
    */
   public Transaction getTransaction()
   {
      return transaction;
   }

   /**
    * Sets the transaction associated with this invocation
    *
    * @param transaction
    */
   public void setTransaction(Transaction transaction)
   {
      this.transaction = transaction;
   }

   /**
    * Retrieves the global transaction associated with this invocation
    *
    * @return the global transaction associated with this invocation
    */
   public GlobalTransaction getGlobalTransaction()
   {
      return globalTransaction;
   }

   /**
    * Sets the global transaction associated with this invocation
    *
    * @param globalTransaction
    */
   public void setGlobalTransaction(GlobalTransaction globalTransaction)
   {
      this.globalTransaction = globalTransaction;
   }

   /**
    * Retrieves the option overrides associated with this invocation
    *
    * @return the option overrides associated with this invocation
    */
   public Option getOptionOverrides()
   {
      if (optionOverrides == null)
      {
         optionOverrides = new Option();
      }
      return optionOverrides;
   }

   public boolean isOptionsUninitialised()
   {
      return optionOverrides == null;
   }

   /**
    * Sets the option overrides associated with this invocation
    *
    * @param optionOverrides
    */
   public void setOptionOverrides(Option optionOverrides)
   {
      this.optionOverrides = optionOverrides;
   }

   /**
    * Tests if this invocation originated locally or from a remote cache.
    *
    * @return true if the invocation originated locally.
    */
   public boolean isOriginLocal()
   {
      return originLocal;
   }

   List<NodeLock> invocationLocks;

   public List<NodeLock> getInvocationLocksAcquired()
   {
      return invocationLocks;
   }

   public void addInvocationLocksAcquired(Collection<NodeLock> locks)
   {
      // no need to worry about concurrency here - a context is only valid for a single thread.
      if (invocationLocks == null) invocationLocks = new ArrayList<NodeLock>(5);
      invocationLocks.addAll(locks);
   }

   public void addInvocationLockAcquired(NodeLock l)
   {
      // no need to worry about concurrency here - a context is only valid for a single thread.
      if (invocationLocks == null) invocationLocks = new ArrayList<NodeLock>(5);
      invocationLocks.add(l);
   }

   public void clearInvocationLocksAcquired()
   {
      invocationLocks = null;
   }

   /**
    * If set to true, the invocation is assumed to have originated locally.  If set to false,
    * assumed to have originated from a remote cache.
    *
    * @param originLocal
    */
   public void setOriginLocal(boolean originLocal)
   {
      this.originLocal = originLocal;
   }

   public String toString()
   {
      return "InvocationContext{" +
            "methodCall=" + methodCall +
            "transaction=" + transaction +
            ", globalTransaction=" + globalTransaction +
            ", optionOverrides=" + optionOverrides +
            ", originLocal=" + originLocal +
            ", txHasMods=" + txHasMods +
            ", cacheLoaderHasMods=" + cacheLoaderHasMods +
            '}';
   }

   public boolean isTxHasMods()
   {
      return txHasMods;
   }

   public void setTxHasMods(boolean b)
   {
      txHasMods = b;
   }

   /**
    * Cache loader might have mods which are different from TX's mods; e.g. when cache is local and passivation is on.
    */
   public boolean isCacheLoaderHasMods()
   {
      return cacheLoaderHasMods;
   }

   public void setCacheLoaderHasMods(boolean cacheLoaderHasMods)
   {
      this.cacheLoaderHasMods = cacheLoaderHasMods;
   }

   public boolean isLocalRollbackOnly()
   {
      return localRollbackOnly;
   }

   /**
    * Resets this to the defaults used when constructing an invocation context object
    */
   public void reset()
   {
      transaction = null;
      globalTransaction = null;
      optionOverrides = null;
      originLocal = true;
      txHasMods = false;
      invocationLocks = null;
   }

   public InvocationContext clone() throws CloneNotSupportedException
   {
      InvocationContext clone = (InvocationContext) super.clone();
      clone.setOptionOverrides(getOptionOverrides().clone());
      return clone;
   }

   /**
    * Sets the state of the InvocationContext based on the template context passed in
    *
    * @param template
    */
   public void setState(InvocationContext template)
   {
      if (template == null)
      {
         throw new NullPointerException("Template InvocationContext passed in to InvocationContext.setState() passed in is null");
      }

      this.setGlobalTransaction(template.getGlobalTransaction());
      this.setLocalRollbackOnly(template.isLocalRollbackOnly());
      this.setOptionOverrides(template.getOptionOverrides());
      this.setOriginLocal(template.isOriginLocal());
      this.setTransaction(template.getTransaction());
      this.setTxHasMods(template.isTxHasMods());
      this.setCacheLoaderHasMods(template.isCacheLoaderHasMods());
   }

   public boolean equals(Object o)
   {
      if (this == o) return true;
      if (o == null || getClass() != o.getClass()) return false;

      final InvocationContext that = (InvocationContext) o;

      if (localRollbackOnly != that.localRollbackOnly) return false;
      if (originLocal != that.originLocal) return false;
      if (txHasMods != that.txHasMods) return false;
      if (cacheLoaderHasMods != that.cacheLoaderHasMods) return false;
      if (globalTransaction != null ? !globalTransaction.equals(that.globalTransaction) : that.globalTransaction != null)
      {
         return false;
      }
      if (optionOverrides != null ? !optionOverrides.equals(that.optionOverrides) : that.optionOverrides != null)
      {
         return false;
      }
      if (transaction != null ? !transaction.equals(that.transaction) : that.transaction != null) return false;

      return true;
   }

   public int hashCode()
   {
      int result;
      result = (transaction != null ? transaction.hashCode() : 0);
      result = 29 * result + (globalTransaction != null ? globalTransaction.hashCode() : 0);
      result = 29 * result + (optionOverrides != null ? optionOverrides.hashCode() : 0);
      result = 29 * result + (originLocal ? 1 : 0);
      result = 29 * result + (txHasMods ? 1 : 0);
      result = 29 * result + (cacheLoaderHasMods ? 1 : 0);
      result = 29 * result + (localRollbackOnly ? 1 : 0);
      return result;
   }

   /**
    * @return the method call associated with this invocation
    */
   public MethodCall getMethodCall()
   {
      return methodCall;
   }

   /**
    * Sets the method call associated with this invocation.
    *
    * @param methodCall methodcall to set
    */
   public void setMethodCall(MethodCall methodCall)
   {
      this.methodCall = methodCall;
   }

   /**
    * Factory method that creates a context with a given method call.
    *
    * @param methodCall methodcall to use
    * @return invocation context
    */
   public static InvocationContext fromMethodCall(MethodCall methodCall)
   {
      InvocationContext ctx = new InvocationContext();
      ctx.methodCall = methodCall;
      return ctx;
   }

   /**
    * If the acq timeout if overwritten for current call, then return that one.
    * If not overwritten return default value.
    */
   public long getContextLockAcquisitionTimeout(long defaultFalue)
   {
      long timeout = defaultFalue;
      if (getOptionOverrides() != null
            && getOptionOverrides().getLockAcquisitionTimeout() >= 0)
      {
         timeout = getOptionOverrides().getLockAcquisitionTimeout();
      }
      return timeout;
   }
}
