/**
 *
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.RPCManager;
import org.jboss.cache.buddyreplication.BuddyManager;
import org.jboss.cache.cluster.ReplicationQueue;
import org.jboss.cache.config.Configuration.CacheMode;
import org.jboss.cache.config.Option;
import org.jboss.cache.factories.annotations.Inject;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionEntry;
import org.jgroups.Address;

import javax.transaction.Transaction;
import java.util.List;

/**
 * Acts as a base for all RPC calls - subclassed by
 * {@link ReplicationInterceptor} and {@link OptimisticReplicationInterceptor}.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public abstract class BaseRpcInterceptor extends MethodDispacherInterceptor
{
   private BuddyManager buddyManager;
   private RPCManager rpcManager;
   private boolean usingBuddyReplication;
   protected boolean defaultSynchronous;
   private ReplicationQueue replicationQueue;

   @Inject
   private void injectComponents(RPCManager rpcManager, BuddyManager buddyManager, ReplicationQueue replicationQueue)
   {
      this.rpcManager = rpcManager;
      this.replicationQueue = replicationQueue;
      this.buddyManager = buddyManager;
      usingBuddyReplication = buddyManager != null && buddyManager.isEnabled();
   }

   @Override
   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      CacheMode mode = cache.getConfiguration().getCacheMode();
      defaultSynchronous = (mode == CacheMode.REPL_SYNC || mode == CacheMode.INVALIDATION_SYNC);
   }

   /**
    * Checks whether any of the responses are exceptions. If yes, re-throws
    * them (as exceptions or runtime exceptions).
    */
   protected void checkResponses(List rsps) throws Throwable
   {
      if (rsps != null)
      {
         for (Object rsp : rsps)
         {
            if (rsp != null && rsp instanceof Throwable)
            {
               // lets print a stack trace first.
               if (log.isDebugEnabled())
                  log.debug("Received Throwable from remote node", (Throwable) rsp);
               throw (Throwable) rsp;
            }
         }
      }
   }

   protected void replicateCall(InvocationContext ctx, MethodCall call, boolean sync, Option o, boolean useOutOfBandMessage) throws Throwable
   {
      replicateCall(ctx, null, call, sync, o, true, useOutOfBandMessage);
   }

   protected void replicateCall(InvocationContext ctx, MethodCall call, boolean sync, Option o) throws Throwable
   {
      replicateCall(ctx, null, call, sync, o);
   }

   protected void replicateCall(InvocationContext ctx, List<Address> recipients, MethodCall call, boolean sync, Option o) throws Throwable
   {
      replicateCall(ctx, recipients, call, sync, o, true, false);
   }

   protected void replicateCall(InvocationContext ctx, List<Address> recipients, MethodCall call, boolean sync, Option o, boolean wrapMethodCallInReplicateMethod, boolean useOutOfBandMessage) throws Throwable
   {

      if (trace) log.trace("Broadcasting call " + call + " to recipient list " + recipients);
      Transaction tx = ctx.getTransaction();
      if (tx != null)
      {
         GlobalTransaction gtx = ctx.getGlobalTransaction();
         TransactionEntry te = cache.getTransactionTable().get(gtx);
         if (te != null)
         {
            if (te.isForceAsyncReplication()) sync = false;
            else if (te.isForceSyncReplication()) sync = true;
         }
      }
      if (!sync && replicationQueue != null && !usingBuddyReplication)
      {
         putCallOnAsyncReplicationQueue(call);
      }
      else
      {
         if (usingBuddyReplication) call = buddyManager.transformFqns(call);

         List<Address> callRecipients = recipients;
         if (callRecipients == null)
         {
            callRecipients = usingBuddyReplication ? buddyManager.getBuddyAddresses() : cache.getMembers();
            if (trace)
               log.trace("Setting call recipients to " + callRecipients + " since the original list of recipients passed in is null.");
         }

         long syncReplTimeout = o.getSyncReplTimeout();
         if (syncReplTimeout < 0) syncReplTimeout = configuration.getSyncReplTimeout();

         MethodCall toCall = wrapMethodCallInReplicateMethod ? toCall = MethodCallFactory.create(MethodDeclarations.replicateMethod_id, call) : call;

         List rsps = rpcManager.callRemoteMethods(callRecipients,
               toCall,
               sync, // is synchronised?
               true, // ignore self?
               (int) syncReplTimeout,
               useOutOfBandMessage
         );
         if (trace) log.trace("responses=" + rsps);
         if (sync) checkResponses(rsps);
      }

   }

   protected void putCallOnAsyncReplicationQueue(MethodCall call)
   {
      if (log.isDebugEnabled()) log.debug("Putting call " + call + " on the replication queue.");
      replicationQueue.add(MethodCallFactory.create(MethodDeclarations.replicateMethod_id, call));
   }

   //todo info expt for this is InvocationContext, move method there
   protected boolean containsModifications(InvocationContext ctx)
   {
      switch (ctx.getMethodCall().getMethodId())
      {
         case MethodDeclarations.prepareMethod_id:
         case MethodDeclarations.optimisticPrepareMethod_id:
            List mods = (List) ctx.getMethodCall().getArgs()[1];
            return mods.size() > 0;
         case MethodDeclarations.commitMethod_id:
         case MethodDeclarations.rollbackMethod_id:
            return ctx.isTxHasMods();
         default:
            return false;
      }
   }

   /**
    * It does not make sense replicating a transaction method(commit, rollback, prepare) if one of the following:
    * <pre>
    *    - call was not initiated here, but on other member of the cluster
    *    - there is no transaction. Why calling a commit if no transaction going on?
    *    - the current transaction did not modufy any data, so other members are not aware of it
    * </pre>
    */
   protected boolean skipReplicationOfTransactionMethod(InvocationContext ctx)
   {
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      boolean isInitiatedHere = gtx != null && !gtx.isRemote();
      if (trace) log.trace("isInitiatedHere? " + isInitiatedHere + "; gtx = " + gtx);
      return !isTransactionalAndLocal(ctx) || !containsModifications(ctx);
   }

   /**
    * The call runs in a transaction and it was initiated on this node of the cluster.
    */
   protected boolean isTransactionalAndLocal(InvocationContext ctx)
   {
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      boolean isInitiatedHere = gtx != null && !gtx.isRemote();
      return isInitiatedHere && (ctx.getTransaction() != null);
   }

   protected boolean isSynchronous(Option option)
   {
      if (option != null)
      {
         if (option.isForceSynchronous())
            return true;
         else if (option.isForceAsynchronous())
            return false;
      }
      return defaultSynchronous;
   }
}