/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 * Created on March 25 2003
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.Region;
import org.jboss.cache.RegionManager;
import org.jboss.cache.eviction.EvictedEventNode;
import org.jboss.cache.eviction.NodeEventType;
import org.jboss.cache.factories.annotations.Inject;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.transaction.GlobalTransaction;

import java.util.Map;

/**
 * Eviction Interceptor.
 * <p/>
 * This interceptor is used to handle eviction events.
 *
 * @author Daniel Huang
 * @author Mircea.Markus@jboss.com
 * @version $Revision: 5104 $
 */
public class EvictionInterceptor extends MethodDispacherInterceptor
{
   protected RegionManager regionManager;

   /**
    * this method is for ease of unit testing. thus package access.
    * <p/>
    * Not to be attempted to be used anywhere else.
    */
   @Inject
   void setRegionManager(RegionManager regionManager)
   {
      this.regionManager = regionManager;
   }

   public EvictionInterceptor()
   {
      initLogger();
   }

   @Override
   protected Object handleEvictVersionedNodeMethod(InvocationContext ctx, Fqn fqn, DataVersion dataVersion) throws Throwable
   {
      return handleEvictMethod(ctx, fqn);
   }

   @Override
   protected Object handleEvictMethod(InvocationContext ctx, Fqn fqn) throws Throwable
   {
      Object retVal = nextInterceptor(ctx);
      // See if the node still exists; i.e. was only data removed
      // because it still has children.
      // If yes, put an ADD event in the queue so the node gets revisited
      boolean complete = (retVal != null && (Boolean) retVal);
      if (!complete)
      {
         if (fqn != null && !canIgnoreEvent(fqn, NodeEventType.ADD_NODE_EVENT))
         {
            registerEvictionEventToRegionManager(new EvictedEventNode(fqn, NodeEventType.ADD_NODE_EVENT, 0), ctx);
         }
      }
      return retVal;
   }

   @Override
   protected Object handlePutKeyValueMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, Object value, boolean createUndoOps) throws Throwable
   {
      return handlePutForExternalReadMethod(ctx, gtx, fqn, key, value);
   }

   @Override
   protected Object handlePutForExternalReadMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, Object key, Object value) throws Throwable
   {
      Object retVal = nextInterceptor(ctx);
      if (fqn != null && key != null && !canIgnoreEvent(fqn, NodeEventType.ADD_ELEMENT_EVENT))
      {
         registerEvictionEventToRegionManager(new EvictedEventNode(fqn, NodeEventType.ADD_ELEMENT_EVENT, 1), ctx);
      }
      return retVal;
   }

   @Override
   protected Object handlePutDataEraseMethod(InvocationContext ctx, GlobalTransaction gt, Fqn fqn, Map newData, boolean createUndoOps, boolean eraseContents) throws Throwable
   {
      Object retVal = nextInterceptor(ctx);
      if (fqn != null && !canIgnoreEvent(fqn, NodeEventType.ADD_NODE_EVENT))
      {
         if (newData == null)
         {
            if (trace)
            {
               log.trace("Putting null data under fqn " + fqn + ".");
            }
         }
         else
         {
            int size;
            synchronized (newData)
            {
               size = newData.size();
            }
            EvictedEventNode event = new EvictedEventNode(fqn, NodeEventType.ADD_NODE_EVENT, size);
            event.setResetElementCount(eraseContents);
            registerEvictionEventToRegionManager(event, ctx);
         }
      }
      return retVal;
   }

   @Override
   protected Object handlePutDataMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, Map data, boolean createUndoOps) throws Throwable
   {
      Object retVal = nextInterceptor(ctx);
      if (fqn != null && !canIgnoreEvent(fqn, NodeEventType.ADD_NODE_EVENT))
      {
         if (data == null)
         {
            if (trace)
            {
               log.trace("Putting null data under fqn " + fqn + ".");
            }

         }
         else
         {
            int size;
            synchronized (data)
            {
               size = data.size();
            }
            registerEvictionEventToRegionManager(new EvictedEventNode(fqn, NodeEventType.ADD_NODE_EVENT, size), ctx);
         }
      }
      return retVal;
   }

   @Override
   protected Object handleRemoveKeyMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, Object key, boolean createUndoOps) throws Throwable
   {
      Object retVal = nextInterceptor(ctx);
      if (retVal == null)
      {
         if (trace)
         {
            log.trace("No event added. Element does not exist");
         }

      }
      else
      {
         if (fqn != null && key != null && !canIgnoreEvent(fqn, NodeEventType.REMOVE_ELEMENT_EVENT))
         {
            registerEvictionEventToRegionManager(new EvictedEventNode(fqn, NodeEventType.REMOVE_ELEMENT_EVENT, 1), ctx);
         }
      }
      return retVal;
   }

   @Override
   protected Object handleGetNodeMethod(InvocationContext ctx, Fqn fqn) throws Throwable
   {
      Object retVal = nextInterceptor(ctx);
      if (retVal == null)
      {
         if (trace)
         {
            log.trace("No event added. Node does not exist");
         }
      }
      else
      {
         if (fqn != null && !canIgnoreEvent(fqn, NodeEventType.VISIT_NODE_EVENT))
         {
            registerEvictionEventToRegionManager(new EvictedEventNode(fqn, NodeEventType.VISIT_NODE_EVENT), ctx);
         }
      }
      return retVal;
   }

   @Override
   protected Object handleGetDataMapMethod(InvocationContext ctx, Fqn fqn) throws Throwable
   {
      return handleGetNodeMethod(ctx, fqn);
   }

   @Override
   protected Object handleGetKeyValueMethod(InvocationContext ctx, Fqn fqn, Object key, boolean sendNodeEvent) throws Throwable
   {
      Object retVal = nextInterceptor(ctx);
      if (retVal == null)
      {
         if (trace)
         {
            log.trace("No event added. Element does not exist");
         }
      }
      else if (fqn != null && key != null && !canIgnoreEvent(fqn, NodeEventType.VISIT_NODE_EVENT))
      {
         registerEvictionEventToRegionManager(new EvictedEventNode(fqn, NodeEventType.VISIT_NODE_EVENT), ctx);
      }
      return retVal;
   }

   @Override
   protected Object handleRemoveNodeMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, boolean createUndoOps) throws Throwable
   {
      Object retVal = nextInterceptor(ctx);
      if (fqn != null && !canIgnoreEvent(fqn, NodeEventType.REMOVE_NODE_EVENT))
      {
         registerEvictionEventToRegionManager(new EvictedEventNode(fqn, NodeEventType.REMOVE_NODE_EVENT), ctx);
      }
      return retVal;
   }

   @Override
   protected Object handleRemoveDataMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, boolean createUndoOps) throws Throwable
   {
      return handleRemoveNodeMethod(ctx, tx, fqn, createUndoOps);
   }

   private void registerEvictionEventToRegionManager(EvictedEventNode event, InvocationContext ctx)
   {
      if (event == null)
      {
         // no node modifications.
         return;
      }

      NodeSPI<?, ?> nodeSPI = peekNode(ctx, event.getFqn(), false, false, false);//cache.peek(event.getFqn(), false);
      //we do not trigger eviction events for resident nodes
      if (nodeSPI != null && nodeSPI.isResident())
      {
         return;
      }

      Region region = regionManager.getRegion(event.getFqn(), false);
      region.putNodeEvent(event);

      if (trace)
      {
         log.trace("Adding event " + event + " to region at " + region.getFqn());
      }

      if (trace)
      {
         log.trace("Finished updating node");
      }
   }

   protected boolean canIgnoreEvent(Fqn fqn, NodeEventType type)
   {
      Region r = regionManager.getRegion(fqn, Region.Type.EVICTION, false);
      // should never happen, we should at least get the default region.
      return r == null || r.getEvictionPolicy().canIgnoreEvent(fqn, type);
   }
}
