package org.jboss.cache.marshall.io;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;

/**
 * An ObjectOutputStream that can safely be reset and reused.  Allows for easy pooling of streams for performance.
 * <p/>
 * This is an ObjectOutputStream with an internal underlying byte stream, which can be queried with {@link #getBytes()}.
 * This stream is reusable - calling {@link #reset()} will ensure the underlying byte stream is reset and the ObjectOutputStream
 * caches are flushed, and headers are written.
 * <p/>
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.1.0
 */
public class ReusableObjectOutputStream extends ObjectOutputStream
{
   ResettableByteArrayOutputStream baos;
   boolean initialised = false;
   private static int defaultBufferSize = 128;

   public ReusableObjectOutputStream() throws IOException, SecurityException
   {
      this(new ResettableByteArrayOutputStream());
   }

   protected ReusableObjectOutputStream(ResettableByteArrayOutputStream baos) throws IOException, SecurityException
   {
      super(baos);
      this.baos = baos;
   }

   /**
    * Initialises a pooled stream for use.
    */
   public void init()
   {
      baos.init();
      initialised = true;
      // write a TC_RESET to any input streams know to clear internal data structures before attempting to read
      baos.write(TC_RESET);
   }

   /**
    * Unlike {@link java.io.ObjectOutputStream#reset()}, this method will flush and wipe the underlying byte stream,
    * and ensure ObjectOutputStream headers are rewritten, awaiting proper reuse.
    *
    * @throws IOException
    */
   @Override
   public void reset() throws IOException
   {
      if (initialised)
      {
         super.reset();
         super.flush();
         baos.reset();
         super.flush();
         baos.trim();
         initialised = false;
      }
   }

   /**
    * @return bytes in the underlying byte stream.  Will internally do a flush first.
    */
   public byte[] getBytes() throws IOException
   {
      flush();
      return baos.toByteArray();
   }

   /**
    * A byte array stream that allows for more direct manipulation of the byte array.
    */
   static class ResettableByteArrayOutputStream extends ByteArrayOutputStream
   {
      ResettableByteArrayOutputStream()
      {
         super(defaultBufferSize);
      }

      int getBufSize()
      {
         return buf.length;
      }

      public void trim()
      {
         // make sure we trim the buffer array to it's original size, to save on memory
         if (buf.length > defaultBufferSize) buf = null;
      }

      public void init()
      {
         if (buf == null) buf = new byte[defaultBufferSize];
         // write the header bytes
         System.arraycopy(ReusableObjectInputStream.INIT_BYTES, 0, buf, 0, 4);
         count = 4;
      }
   }
}
