package org.jboss.cache.transaction;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.RPCManager;
import org.jboss.cache.RPCManagerImpl;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.factories.ComponentRegistry;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.misc.TestingUtil;
import org.jgroups.Address;
import org.jgroups.blocks.RspFilter;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import javax.transaction.TransactionManager;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.CountDownLatch;

/**
 * This is to test the scenario described in http://jira.jboss.org/jira/browse/JBCACHE-1270
 * <p/>
 * i) Node A sends prepare for GTX1; synchronous. Gets applied on Node B. Locks are held on B.
 * ii) Node A sends commit for GTX1; *asynchronous*.
 * iii) Node A sends lots of other messages related to other sessions.
 * iv) Node A sends prepare for GTX2; synchronous.
 * v) Node B is busy, and by luck the GTX2 prepare gets to UNICAST before the GTX1 commit.
 * vi) GTX2 prepare blocks due to locks from GTX1.
 * vii) GTX1 commit is blocked in UNICAST because another thread from Node A is executing.
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.1.0
 */
@Test(groups = "functional")
public class PrepareCommitContentionTest
{
   CacheSPI<Object, Object> c1, c2;

   @BeforeMethod
   public void setUp() throws CloneNotSupportedException
   {
      c1 = (CacheSPI<Object, Object>) new DefaultCacheFactory<Object, Object>().createCache(false);
      c1.getConfiguration().setCacheMode(Configuration.CacheMode.REPL_SYNC);
      c1.getConfiguration().setTransactionManagerLookupClass(DummyTransactionManagerLookup.class.getName());
      c1.getConfiguration().setLockAcquisitionTimeout(5000);
      c2 = (CacheSPI<Object, Object>) new DefaultCacheFactory<Object, Object>().createCache(c1.getConfiguration().clone(), false);
   }

   @AfterMethod
   public void tearDown()
   {
      TestingUtil.killCaches(c1, c2);
   }

   public void testWithSyncCommitPhase() throws Exception
   {
      doTest(true, false);
   }

   public void testWithDefautCommitPhase() throws Exception
   {
      doTest(false, false);
   }

   public void testControl() throws Exception
   {
      try
      {
         doTest(false, true);
         assert false : "Should fail if we don't use out of band messages for non-sync commits";
      }
      catch (AssertionError expected)
      {
         // should fail
      }
   }

   private void doTest(final boolean syncCommit, boolean noOutOfBandMessages) throws Exception
   {
      c1.getConfiguration().setSyncCommitPhase(syncCommit);
      c2.getConfiguration().setSyncCommitPhase(syncCommit);

      final CountDownLatch mainThreadCommitLatch = new CountDownLatch(1);
      final CountDownLatch secondThreadPrepareLatch = new CountDownLatch(1);
      final Fqn fqn = Fqn.fromString("/a/b/c");

      DelayingRPCManager delayingRPCManager = new DelayingRPCManager(mainThreadCommitLatch, secondThreadPrepareLatch, syncCommit, noOutOfBandMessages);
      ComponentRegistry cr = TestingUtil.extractComponentRegistry(c1);
      cr.registerComponent(RPCManager.class.getName(), delayingRPCManager, RPCManager.class);

      c1.start();
      c2.start();

      TestingUtil.blockUntilViewsReceived(60000, c1, c2);

      TransactionManager tm = c1.getTransactionManager();

      Thread secondTransaction = new Thread("SecondThread")
      {
         public void run()
         {
            // wait until thread1 finishes the prepare.
            try
            {
               secondThreadPrepareLatch.await();
            }
            catch (InterruptedException e)
            {
               // do nothing
            }

            try
            {
               // now replicate a put on the SAME node so there is lock contention
               TransactionManager tm = c1.getTransactionManager();
               if (syncCommit)
                  mainThreadCommitLatch.countDown(); // we need to release the main thread commit latch first otherwise it will deadlock!

               tm.begin();
               c1.put(fqn, "k", "v2");
               tm.commit();
            }
            catch (Exception e)
            {
               e.printStackTrace();
            }
         }
      };
      secondTransaction.start();


      tm.begin();
      c1.put(fqn, "k", "v");
      tm.commit();

      secondTransaction.join();

      // now assert that both transactions have succeeded
      assert c1.get(fqn, "k").equals("v2");
      assert c2.get(fqn, "k").equals("v2");
   }

   public static class DelayingRPCManager extends RPCManagerImpl
   {
      CountDownLatch mainThreadLatch, secondThreadLatch;
      boolean syncCommit;
      boolean noOOBMessages = false;

      public DelayingRPCManager(CountDownLatch latch1, CountDownLatch latch2, boolean syncCommit, boolean noOOBMessages)
      {
         mainThreadLatch = latch1;
         secondThreadLatch = latch2;
         this.syncCommit = syncCommit;
         this.noOOBMessages = noOOBMessages;
      }

      @Override
      public List<Object> callRemoteMethods(final List<Address> recipients, final MethodCall methodCall, final int mode, final boolean excludeSelf, final long timeout, final RspFilter responseFilter, final boolean oob) throws Exception
      {
         if (isPrepareMethod(methodCall) && Thread.currentThread().getName().equals("SecondThread"))
         {
            if (!syncCommit) mainThreadLatch.countDown();
         }
         else if (isCommitMethod(methodCall) && !Thread.currentThread().getName().equals("SecondThread"))
         {
            Thread th = new Thread()
            {
               public void run()
               {
                  try
                  {
                     secondThreadLatch.countDown(); // let the secondTransaction start it's prepapre
                     mainThreadLatch.await(); // and block arbitrarily until the secondTransaction informs us to proceed
                     Thread.sleep(1000);
                     DelayingRPCManager.super.callRemoteMethods(recipients, methodCall, mode, excludeSelf, timeout, responseFilter, !noOOBMessages && oob);
                  }
                  catch (Exception e)
                  {
                     e.printStackTrace();
                  }
               }
            };
            th.start();
            if (syncCommit) th.join();

            return Collections.emptyList();
         }
         return super.callRemoteMethods(recipients, methodCall, mode, excludeSelf, timeout, responseFilter, !noOOBMessages && oob);
      }

      private boolean isCommitMethod(MethodCall call)
      {
         return call.getMethodId() == MethodDeclarations.commitMethod_id ||
               (call.getMethodId() == MethodDeclarations.replicateMethod_id && isCommitMethod((MethodCall) call.getArgs()[0]));
      }

      private boolean isPrepareMethod(MethodCall call)
      {
         return call.getMethodId() == MethodDeclarations.prepareMethod_id ||
               (call.getMethodId() == MethodDeclarations.replicateMethod_id && isPrepareMethod((MethodCall) call.getArgs()[0]));
      }
   }
}
